from act.scio.plugin import BasePlugin, Result
import addict
from typing import Text, List, Dict, Set
import ipaddress
import re


class Plugin(BasePlugin):
    name = "vulnerabilities"
    info = "Extracting Vulnerability References (CVE, MSID) from body of text"
    version = "0.1"
    dependencies: List[Text] = []

    cve = re.compile("\\b(?:CVE|cve)-\d{4}-\d{4,7}\\b")
    msid = re.compile("\\b(?:MS|ms)\d{2}-\d+\\b")

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        text = nlpdata.content

        res = addict.Dict()

        res.cve = self.cve.findall(text)
        res.msid = self.msid.findall(text)

        return Result(name=self.name, version=self.version, result=res)




"""All functions related to scio upload"""

from typing import IO, Text, Dict
from logging import warning

import base64
import time

import requests


def read_as_base64(obj: IO) -> Text:
    """Create a base64 encoded string from a file like object"""

    encoded_bytes = base64.b64encode(obj.read())
    return encoded_bytes.decode('ascii')


def to_scio_submit_post_data(obj: IO, filemap: Dict) -> Dict[Text, Text]:
    """Take a file like object, and return a dictionary on the correct form for
    submitting to the SCIO API (https://github.com/mnemonic-no/act-scio-api)"""

    metadata = filemap.copy()
    metadata['content'] = read_as_base64(obj)

    assert 'content' in metadata
    assert 'filename' in metadata
    assert 'uri' in metadata

    return metadata


def upload(url: Text, filemap: Dict) -> None:
    """Upload a file to the Scio engine"""

    retry_codes = [429, None]

    with open(filemap['filename'], "rb") as file_h:
        post_data = to_scio_submit_post_data(file_h, filemap)

        # Disable all proxy settings from environment
        session = requests.Session()
        session.trust_env = False

        status_code = None
        while status_code in retry_codes:
            req = session.post(url, json=post_data)
            status_code = req.status_code

            if status_code not in retry_codes and status_code != 200:
                raise UploadError("Status {0}".format(req.status_code))

            if status_code in retry_codes:
                warning("%s: %s", status_code, req.text)
                time.sleep(1)


class UploadError(Exception):
    """Wrap all errors related to uploading data to scio"""



import addict
import nltk
from act.scio.plugin import BasePlugin, Result
from typing import Text, List


class Plugin(BasePlugin):
    name = "pos_tag"
    info = "Part-of-speach tagging of a body of text"
    version = "0.1"
    dependencies: List[Text] = []

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        tokens = nltk.word_tokenize(nlpdata.content)

        res = addict.Dict()

        res.tokens = nltk.pos_tag(tokens)

        return Result(name=self.name, version=self.version, result=res)



"""
Helper functions for retrieving vocabulary aliases
"""

import argparse
import re
from typing import Any, Dict, List, Optional, Tuple

import requests
import urllib3  # type: ignore

ESCAPE_CHARS = [",", ":", "#"]


def unescape(s: str) -> str:
    """
    Unescape
    """

    for char in ESCAPE_CHARS:
        s = s.replace("\\{}".format(char), char)

    return s


def escape(s: str) -> str:
    """
    Escape
    """
    for char in ESCAPE_CHARS:
        s = s.replace(char, "\\{}".format(char))

    return s


class UnknownResult(Exception):
    """UnknownResult is used in API request (not 200 result)"""

    def __init__(self, *args: Any, **kwargs: Dict[Any, Any]) -> None:
        Exception.__init__(self, *args, **kwargs)


def parse_aliases(line: str) -> Tuple[str, List[str]]:
    """
    Parse line on this form:

    name:alias1, alias2, alias3

    Supports escaping of ":", "," and "#"

    returns:
    name and list of aliases
    """

    # Split on ":" unless they are escaped
    vocab, aliases_str = re.split(r'(?<!\\):', line, maxsplit=1)

    # Split aliases on ",", unless they are escaped
    aliases = [unescape(a.strip())
               for a in re.split(r'(?<!\\),', aliases_str)
               if a.strip()]

    return unescape(vocab.strip()), aliases


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = argparse.ArgumentParser(description="Fetch ISO 3166 data")
    parser.add_argument('--http-timeout', dest='timeout', type=int,
                        default=120, help="Timeout")
    parser.add_argument(
        '--proxy-string',
        dest='proxy_string',
        help="Proxy to use for external queries")
    parser.add_argument(
        '--data-dir',
        dest='data_dir',
        required=True,
        help="Output path for data")

    return parser


def output_alias(filename: str, alias_map: Dict, exclude: Optional[List] = None) -> None:
    """
    Output alias to file
    """
    if not exclude:
        exclude = []

    with open(filename, "w") as f:
        for key in sorted(alias_map.keys()):
            aliases = alias_map[key]
            aliases = [escape(alias).strip() for alias in sorted(aliases)
                       if alias not in exclude and alias != key]
            f.write("{}:{}\n".format(escape(key).strip(), ",".join(aliases)))


def merge(*lists: Dict, lower: bool = False) -> Dict:
    """
    Combine multiple alias lists in order.
    If key exists as one of the previous alias lists, add aliases to the previous
    main key
    """

    combined: Dict = {}

    for alias_list in lists:
        for key, aliases in alias_list.items():
            if lower:
                key = key.lower()
                aliases = [alias.lower() for alias in aliases]

            # One of our aliases is already main group name in ATT&CK.
            # Merge aliases from MISP (including "main" group name with ATT&CK)
            alias_in_keys = [alias for alias in aliases if combined.get(alias)]

            if alias_in_keys:
                key = alias_in_keys[0]

            combined[key] = list(set(combined.get(key, []) + aliases))

    return combined


def fetch_json(
        url: str,
        proxy_string: str = "",
        timeout: int = 60,
        verify_https: bool = True) -> Dict:
    """Fetch remote URL as JSON
    url (string):                    URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """

    proxies = {
        'http': proxy_string,
        'https': proxy_string
    }

    options = {
        "verify": verify_https,
        "timeout": timeout,
        "proxies": proxies,
        "params": {}
    }

    if not verify_https:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    # Unable to infer type
    req = requests.get(url, **options)  # type: ignore

    if not req.status_code == 200:
        errmsg = "status_code: {0.status_code}: {0.content}"
        raise UnknownResult(errmsg.format(req))

    # Unable to infer type of json()
    return req.json()  # type: ignore



"""Helper function related to feed configuration"""

import os
from enum import Enum
from typing import Text, Optional, Tuple, List
import argparse
import caep
import logging


FeedType = Enum('FeedType', ['none', 'partial', 'full'])

XDG_CONFIG = os.path.expanduser(os.environ.get("XDG_CONFIG_HOME", "~/.config"))
XDG_CACHE = os.path.expanduser(os.environ.get("XDG_CACHE_HOME", "~/.cache"))

def get_args() -> argparse.Namespace:
    """initialize argument parser"""

    parser = argparse.ArgumentParser(description="pull feeds into html files")
    parser.add_argument("-l", "--log", type=str,
                        help="Which file to log to (default: stdout)")
    parser.add_argument("--file-format", nargs="+", default="pdf doc xls csv xml")
    parser.add_argument(
        "--store-path",
        help=f"Location for stored files. Default {XDG_CACHE}/scio-feeds")
    parser.add_argument('--proxy-string', help="Proxy to use for external queries")
    parser.add_argument("--ignore", type=str, help="file with ignore patterns")
    parser.add_argument("--feeds", default=caep.get_config_dir("scio/etc/feeds.txt"),
                        type=str, help=f"feed urls (one pr. line). Default: {XDG_CONFIG}/scio/etc/feeds.txt")
    parser.add_argument("--cache", default=caep.get_cache_dir("scio-feeds/cache.db"),
                        help=f"sqlite db containing cached hashes. Default = {XDG_CACHE}/scio-feeds/cache.db")
    parser.add_argument("--scio", help="Upload to scio engine API url. " +
                        "Set to empty value to not upload files.",
                        default="http://localhost:3000/submit")
    parser.add_argument("--stoplist", default=caep.get_config_dir("scio/etc/secstoplist.txt"),
                        help="Provided own stoplist for text extraction")
    parser.add_argument('--logfile')
    parser.add_argument('--loglevel', default="info")

    args: argparse.Namespace = caep.config.handle_args(parser, "scio/etc", "scio.ini", "feeds")

    return args


def get_feed_url_from_feed_file_line(line: Text) -> Optional[Text]:
    """Extract the url portion from the feed file line"""

    if len(line) <= 2:
        logging.warning("Missconfigured feed file line: %s", line)
        return None

    feed_url = line[2:].strip()
    if not feed_url:
        logging.warning("Empty URL part of feed file line: %s", line)

    return feed_url


def get_feed_kind(line: Text) -> FeedType:
    """Extract the feed kind from a feed config line"""

    if len(line) < 2:
        return FeedType.none
    type_text = line[:2]
    if type_text == 'p ':
        return FeedType.partial
    if type_text == 'f ':
        return FeedType.full

    return FeedType.none


def parse_feed_file(filename: Text) -> Tuple[List[Text], List[Text]]:
    """Parse feed file, split feeds into partial and full feeds
    (lines starting with 'f ' and 'p ')"""

    full_feeds = []
    partial_feeds = []

    for linenum, feed_line in enumerate(open(os.path.expanduser(filename))):
        feed_type = get_feed_kind(feed_line)
        feed_url = get_feed_url_from_feed_file_line(feed_line)
        if not feed_url:
            continue

        if feed_type == FeedType.none:
            logging.error("Unable to parse line %s [%s]", linenum + 1, feed_line)
            continue

        if feed_type == FeedType.full:
            full_feeds.append(feed_url)
        if feed_type == FeedType.partial:
            partial_feeds.append(feed_url)

    return full_feeds, partial_feeds



#!/usr/bin/env python3
"""Copyright 2019,2020 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

---
Download feeds from the feeds.txt file, store feed content in .html and feed
metadata in .meta files. Also attempts to download links to certain document
types"""

from typing import List, Text, Dict

import datetime
import hashlib
import logging
import os
import urllib3

from act.scio.config import get_cache_dir
from act.scio.feeds import conf, download, cache, upload
from act.scio.logsetup import setup_logging

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def sha256_of_file(filename: Text) -> Text:
    """Compute sha256 hexdigest of file"""

    with open(filename, "rb") as f:
        data = f.read()
        return hashlib.sha256(data).hexdigest()


def upload_uncached_files(cache_file: Text, files: List[Dict], scio_url: Text) -> int:
    """Check each downloaded file hexdigest against a cache of previously uploaded
    files. Only upload "new" files."""

    mycache = cache.Cache(cache_file)

    nup = 0

    for filemap in files:

        filename: Text = filemap['filename']
        sha256 = sha256_of_file(filename)

        if not mycache.contains(sha256):
            try:
                if scio_url != "dummy.url":
                    upload.upload(scio_url, filemap)
                mycache.insert(filename, sha256, str(datetime.datetime.now()))
                logging.info("Uploaded %s to scio", filename)
                nup += 1
            except upload.UploadError as err:
                logging.error(err)

    return nup


def main() -> None:
    """Main program loop. entry point"""

    args = conf.get_args()

    if not args.store_path:
        args.store_path = get_cache_dir("scio-feeds", create=True)

    p = os.path.join(args.store_path, "download")
    if not os.path.isdir(p):
        os.makedirs(p)

    setup_logging(args.loglevel, args.logfile, "scio-feed-download")

    files: List[Dict] = []

    try:
        full_feeds, partial_feeds = conf.parse_feed_file(args.feeds)
        files += download.download_feed_list(args, full_feeds, partial=False)
        files += download.download_feed_list(args, partial_feeds, partial=True)
    except IOError as err:
        logging.error(str(err))
        raise err

    if not args.scio:
        logging.info("No Scio API Url provided. Exit after download[%s]", len(files))
        return

    logging.info("Checking upload status of %s files", len(files))

    nup = upload_uncached_files(args.cache, files, args.scio)

    logging.info("Uploaded %s files", nup)


if __name__ == "__main__":
    main()



"""Setup script for the python-act library module"""

from os import path

from setuptools import setup

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), "rb") as f:
    long_description = f.read().decode('utf-8')

setup(
    name="act-scio",
    version="0.0.27",
    author="mnemonic AS",
    zip_safe=True,
    author_email="opensource@mnemonic.no",
    description="ACT SCIO",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT",
    keywords="ACT, mnemonic",
    entry_points={
        'console_scripts': [
            'scio-analyze = act.scio.analyze:main',
            'scio-api = act.scio.api:main',
            'scio-config = act.scio.scio_config:main',
            'scio-feeds = act.scio.feeds.feeds:main',
            'scio-tika-server = act.scio.tika_engine:main',
            'scio-nltk-download= act.scio.nltk_download:main',
            'scio-upload = act.scio.upload:main',
        ]
    },

    # Include ini-file(s) from act/workers/etc
    package_data={'act.scio': ['etc/*', 'etc/plugins/*', 'vendor/*']},
    packages=["act.scio", "act.scio.plugins", "act.scio.feeds"],

    # https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
    # __init__.py under all packages under in the act namespace must contain exactly string:
    # __path__ = __import__('pkgutil').extend_path(__path__, __name__)
    namespace_packages=['act'],
    url="https://github.com/mnemonic-no/act-scio2",
    install_requires=[
        "addict",
        "aiofiles",
        "beautifulsoup4",
        "bs4",
        "caep",
        "elasticsearch",
        "elasticsearch_dsl",
        "fastapi",
        "feedparser",
        "greenstalk>=2.0.0",
        "ipaddress",
        "justext",
        "nltk",
        "pytest-asyncio",
        "pytest",
        "python-magic",
        "pytz",
        "requests",
        "tika",
        "urllib3",
        "uvicorn",
    ],
    python_requires='>=3.6, <4',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: ISC License (ISCL)",
    ],
)



import addict
from act.scio.vocabulary import Vocabulary
from act.scio.plugin import BasePlugin, Result
from typing import Text, List
import configparser
import os.path


class Plugin(BasePlugin):
    name = "tools"
    info = "Extracting references to known tools from a body of text"
    version = "0.2"
    dependencies: List[Text] = []

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        ini = configparser.ConfigParser()
        ini.read([os.path.join(self.configdir, "tools_pattern.ini")])
        ini['tools']['alias'] = os.path.join(self.configdir, ini['tools']['alias'])

        vocab = Vocabulary(ini['tools'])

        res = addict.Dict()

        res.Tools = vocab.regex_search(nlpdata.content, debug=self.debug)

        return Result(name=self.name, version=self.version, result=res)



"""Copyright 2019,2020 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

---
Program to check the feed output directory of feed_downloader.py
Any downloaded content not found in the cache is handled.
The .meta files associated with the .html files from the feed download is
kept to preserve information about source and original titles as well as
feed time and date stamps.
This files builds a document understood by SCIOs 'doc' work queue.
This utility is not meant to handle the files downloaded from links _in_ the
feed as these document does not share any of the meta-data associated with
the feed entries. These files need to be handled separately (using SCIOs own
submit utility.

---
Helper functions used for making desitions and modifications
"""

from typing import List, Text, Optional
import logging
import os
import urllib
import urllib.parse


def parse_and_correct_link(link: Text) -> urllib.parse.ParseResult:
    """Parse the link and rewrites known web-view vs raw store locations (e.g. github)"""

    parsed = urllib.parse.urlparse(link)

    if parsed.netloc == 'github.com':
        parsed = parsed._replace(netloc="raw.githubusercontent.com")
        parsed = parsed._replace(path=parsed.path.replace("/blob/", "/", 1))
        logging.info("found github link. Modify to get raw content")

    return parsed


def in_ignore_file(fname: Text, ignore_file: Optional[Text]) -> bool:
    """Check if a spesific filename is present in the ignore file (if any)"""

    if not ignore_file:
        return False

    ignore_file = os.path.expanduser(ignore_file)

    if not os.path.isfile(ignore_file):
        logging.warning("Ignore file not found: %s", ignore_file)
        return False

    with open(ignore_file) as f:
        ignored = [line.strip() for line in f]
        if fname in ignored:
            return True

    return False


def extract_file_extension(url: urllib.parse.ParseResult) -> Text:
    """Extract the extension (if any) from the path.
    If there are no extension, return the empty string."""

    file_extension = url.path.rsplit(".", 1)
    if len(file_extension) == 2:
        return file_extension[1]
    return ""


def filter_links(file_formats: List[Text],
                 links: List[urllib.parse.ParseResult]) -> List[urllib.parse.ParseResult]:
    """Run though a list of urls, checking if they contains certain
    elements that looks like possible file download possibilities"""

    file_format_set = set(file_formats)

    def __contains_fileextension(url: urllib.parse.ParseResult) -> bool:
        ext = extract_file_extension(url)
        return ext in file_format_set

    return list(filter(__contains_fileextension, links))
